from typing import List
from curses import wrapper, error

from cronotrack.services.ui_manger import UIManager
from cronotrack.services.chrono_timer import ChronoTimer


class App(object):
    def run(self):
        chrono_list = []
        ui_manager = UIManager(chrono_list)

        ct = ChronoTimer()
        ct.start()
        chrono_list.append(ct)

        try:
            wrapper(ui_manager.run)
        except error as e:
            print("Application finished with unexpected error: \n> > {}\n".format(str(e)))

        self._print_summary(chrono_list)

    def _print_summary(self, chrono_list: List[ChronoTimer]):
        print("\n{:^54s}".format('SUMMARY'))
        print("{}".format('-'*54))
        for chrono in chrono_list:
            l = '[{:19s}] - [{:19s}] {}'.format(chrono.get_str_dt_start(),
                                                chrono.get_str_dt_end(),
                                                chrono.get_str_current_crono())
            if chrono.name:
                l += ' | {}'.format(chrono.name)

            print(l)

        print()



import curses, curses.textpad
from collections import namedtuple

from .chrono_timer import ChronoTimer


class UIWorker():
    """
    Class to interact with terminal UI using curses library
    """
    _COLOR_BG_CYAN = 1
    _COLOR_BG_GREEN = 2
    _COLOR_BG_YELLOW = 3
    _COLOR_BG_RED = 4
    _COLOR_BG_WHITE = 5

    def __init__(self, screen: curses.initscr()):
        self._screen = screen
        self._cursor_pos = Position(0, 0)
        self._init_colors()

    @staticmethod
    def get_init_instance():
        ui_worker = UIWorker(None)
        curses.wrapper(ui_worker._init_screen)

        return ui_worker

    def write_chrono_line(self, chrono_timer: ChronoTimer, end_line=False):
        # self._screen.move(self._cursor_pos.y, self._cursor_pos.x)
        pos_x = self._cursor_pos.x
        pos_y = self._cursor_pos.y

        eol = '\n' if end_line else ''
        chrono_style = self._get_crono_style(chrono_timer)

        output = ' [{}] '.format(chrono_timer.get_str_dt_start())
        pos_x, _ = self._write_line(pos_y, pos_x, output, curses.A_BOLD)

        output = '{}'.format(chrono_timer.get_str_current_crono())
        pos_x, _ = self._write_line(pos_y, pos_x, output, chrono_style)

        if chrono_timer.name:
            output = ' - {}'.format(chrono_timer.name)
            pos_x, _ = self._write_line_with_limit(pos_y, pos_x, output)

        self._write_line(pos_y, pos_x, eol)
        self._screen.refresh()

        if end_line:
            self._cursor_pos.y += 1

    def show_help_bar(self, pos_y=-1, pos_x=3):
        if pos_y == -1:
            max_y, max_x = self._screen.getmaxyx()
            pos_y = max_y - 1

        help_str = '|q: quit, p: pause/resume, n: new_chron'
        output = '{}{}'.format(help_str, ' '*(max_x - len(help_str) - pos_x - 1))
        self._screen.addstr(pos_y, pos_x, output, curses.color_pair(self._COLOR_BG_WHITE))

        self._screen.refresh()

    def write_keystroke(self, key: str, pos_y=-1, pos_x=0):
        if pos_y == -1:
            pos_y, _ = self._screen.getmaxyx()

        self._screen.addstr(pos_y-1, pos_x, ':{}'.format(key))
        self._screen.refresh()

    def _write_line(self, pos_y: int, pos_x: int, output: str, style=curses.A_NORMAL) -> (int, int):
        self._screen.addstr(pos_y, pos_x, output, style)

        new_x = pos_x + len(output)
        new_y = pos_y + 1 if output.endswith('\n') else pos_y

        return (new_x, new_y)

    def _write_line_with_limit(self, pos_y: int, pos_x: int, output: str, limit=-1, style=curses.A_NORMAL) -> (int, int):
        """
        Write a string until a limit is reached. If limit is -1 (default) the limit woluld be the current line
        """
        if limit == -1:
            _, max_x = self._screen.getmaxyx()
            limit = max_x - pos_x - 1

        self._screen.addnstr(pos_y, pos_x, output, limit, style)

        new_x = pos_x + min(len(output), limit)
        new_y = pos_y + 1 if output.endswith('\n') else pos_y

        return (new_x, new_y)

    def show_text_input(self, pos_y=-1, pos_x=1, add_frame=True, placeholder="") -> str:
        max_y, max_x = self._screen.getmaxyx()
        if pos_y == -1:
            pos_y = max_y - 3

        if add_frame:
            curses.textpad.rectangle(self._screen,
                                     pos_y - 1, 0,
                                     pos_y + 1, max_x - 1)

        self._screen.addstr(pos_y, pos_x, placeholder)

        txt_win = curses.newwin(1, max_x - len(placeholder),
                                pos_y, pos_x + len(placeholder))

        txt_box = curses.textpad.Textbox(txt_win)
        curses.curs_set(1)
        self._screen.refresh()

        txt_input = txt_box.edit()
        self._clear_lines([pos_y -1, pos_y, pos_y + 1])
        curses.curs_set(0)

        return txt_input

    def _clear_lines(self, lines_indices):
        for line_index in lines_indices:
            self._screen.move(line_index, 0)
            self._screen.clrtoeol()

    def _init_colors(self):
        curses.start_color()
        curses.init_pair(self._COLOR_BG_CYAN, curses.COLOR_BLACK, curses.COLOR_CYAN)
        curses.init_pair(self._COLOR_BG_GREEN, curses.COLOR_BLACK, curses.COLOR_GREEN)
        curses.init_pair(self._COLOR_BG_YELLOW, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        curses.init_pair(self._COLOR_BG_RED, curses.COLOR_WHITE, curses.COLOR_RED)
        curses.init_pair(self._COLOR_BG_WHITE, curses.COLOR_BLACK, curses.COLOR_WHITE)

    def _init_screen(self, screen):
        self._screen = screen

    def _get_crono_style(self, chrono_timer: ChronoTimer):
        style = curses.color_pair(self._COLOR_BG_CYAN)
        if chrono_timer.is_stopped:
            style = curses.color_pair(self._COLOR_BG_GREEN)
        elif chrono_timer.is_paused:
            style = curses.color_pair(self._COLOR_BG_RED) | curses.A_BLINK

        return style


class Position():
    def __init__(self, x, y):
        self.x = x
        self.y = y


import curses
from typing import List

from .ui_worker import UIWorker
from .chrono_timer import ChronoTimer


class UIManager():
    def __init__(self, chrono_list=None):
        self.chrono_list: List[ChronoTimer] = chrono_list if chrono_list is not None else []

        self._ui_worker: UIWorker = None
        self._current_chr_index: int = 0
        self._last_keystroke = -1
        self._is_init = False

    def run(self, screen):
        self._init(screen)

        try:
            while True:
                self._render_stopped_chronos()
                self._render_current_chronos()
                self._ui_worker.show_help_bar()
                self._handle_keystrokes(screen)
        except KeyboardInterrupt:
            pass

    def _render_stopped_chronos(self):
        """
        Assumption: Stoped chronos always will be at the beginig of the list.
            So when no stopped chrono found the rest should be also not stopped
        """
        if len(self.chrono_list) == 0 or self._current_chr_index == len(self.chrono_list):
            return
        
        last_stopped_index = self._current_chr_index
        for i in range(self._current_chr_index, len(self.chrono_list)):
            chrono = self.chrono_list[i]
            if chrono.is_stopped:
                self._ui_worker.write_chrono_line(chrono, end_line=True)
                last_stopped_index = i + 1

        self._current_chr_index = last_stopped_index

    def _render_current_chronos(self):
        """
        Assumption: There should be at most only one chrono not stopped(currently used).
            If this chrono exists it will be at the end of the list
        """
        if len(self.chrono_list) == 0:
            return

        last_chrono = self.chrono_list[-1]
        if not last_chrono.is_stopped:
            self._ui_worker.write_chrono_line(last_chrono)

    def _handle_keystrokes(self, screen: curses.initscr()):
        try:
            key = screen.getkey()
            self._ui_worker.write_keystroke(key+' ')
            
            if key == 'q':
                self._handle_quit()
            elif key == 'p':
                self._handle_pause()
            elif key == 'n':
                self._handle_new_chrono()

        except curses.error:
            pass

    def _handle_pause(self):
        if len(self.chrono_list) == 0:
            return

        last_chrono = self.chrono_list[-1]
        if last_chrono.is_stopped:
            return

        if last_chrono.is_paused:
            last_chrono.start()
        else:
            last_chrono.pause()

    def _handle_stop(self):
        if len(self.chrono_list) == 0:
            return

        last_chrono = self.chrono_list[-1]
        if last_chrono.is_stopped:
            return

        # Add name to previous crono
        name_input = self._ui_worker.show_text_input(placeholder="Name: ")
        last_chrono.name = name_input

        last_chrono.stop()

    def _handle_new_chrono(self):
        self._handle_stop()

        new_chrono = ChronoTimer()
        new_chrono.start()
        self.chrono_list.append(new_chrono)

    def _handle_quit(self):
        self._handle_stop()
        raise KeyboardInterrupt('Stopped')

    def _init(self, screen: curses.initscr()):
        if not self._is_init:
            curses.curs_set(0)
            curses.noecho()
            screen.nodelay(1)

            self._ui_worker = UIWorker(screen)
            self._is_init = True


# -*- encoding: utf-8 -*-
from setuptools import setup, find_packages
from pathlib import Path
import re

# Get current version
version = '0.1.0'
version_input_path = Path('./setup/version.txt')
if version_input_path.exists():
    with open(version_input_path, 'r') as f:
        version_input = f.readline().rstrip()

    if re.match('\\d+\\.\\d+\\.\\d+', version_input):
        version = version_input

# Load doc
with open('README.md') as f:
    readme = f.read()

setup(
    name='cronotrack',
    version=version,
    description='CLI App to track time',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='salpreh',
    author_email='salva.perez46@gmail.com',
    url='https://github.com/salpreh/cronotrack',
    license='GPL-3.0',
    packages=find_packages(exclude=('test', 'assets', 'dev_test', 'doc'))
)



from datetime import datetime
import time


class ChronoTimer():
    """
    Class to manage a chronometer execution
    """
    DATETIME_FMT = '%Y-%m-%d %H:%M:%S'
    TIME_FMT = '%H:%M:%S'

    def __init__(self):
        self.name = ''
        self.dt_start = None
        self.dt_end = None
        self.ts_start = []
        self.ts_end = []

        self._ts_current = 0
        self._last_ts_start = None
        self._stopped = False

    def __eq__(self, other):
        self_val = self._get_compare_val(self)
        other_val = self._get_compare_val(other)

        return self_val == other_val

    def start(self):
        if not self.dt_start:
            self.dt_start = datetime.now()

        ts_start = time.perf_counter()

        self.ts_start.append(ts_start)
        self._update_current_ts(ts_start=ts_start)

    def stop(self):
        self.dt_end = datetime.now()
        self._stopped = True
        if len(self.ts_end) < len(self.ts_start):
            ts_end = time.perf_counter()
            self.ts_end.append(ts_end)
            self._update_current_ts(ts_end=ts_end)

    def pause(self):
        if len(self.ts_end) == len(self.ts_start):
            raise ValueError('Timer already paused or stopped')

        ts_end = time.perf_counter()
        self.ts_end.append(ts_end)
        self._update_current_ts(ts_end=ts_end)

    def get_current_crono(self):
        if self._last_ts_start:
            return self._ts_current + time.perf_counter() - self._last_ts_start

        return self._ts_current

    def get_str_dt_start(self):
        if not self.dt_start:
            return ''

        return datetime.strftime(self.dt_start, ChronoTimer.DATETIME_FMT)

    def get_str_dt_end(self):
        if not self.dt_end:
            return ''

        return datetime.strftime(self.dt_end, ChronoTimer.DATETIME_FMT)

    def get_str_current_crono(self):
        return time.strftime(ChronoTimer.TIME_FMT, time.gmtime(self.get_current_crono()))

    @property
    def is_stopped(self):
        return self._stopped

    @property
    def is_paused(self):
        return len(self.ts_end) == len(self.ts_start)

    def _update_current_ts(self, ts_end=None, ts_start=None):

        if ts_start and not self._last_ts_start:
            self._last_ts_start = ts_start
        elif ts_end:
            self._ts_current += ts_end - self._last_ts_start
            self._last_ts_start = None
        else:
            raise ValueError('Start time incoherece when updating crono')

    def _refresh_current_ts(self):
        self._ts_current = 0
        index = 0
        for index, (ts_end, ts_start) in zip(self.ts_end, self.ts_start):
            self._ts_current += ts_end + ts_start

        if len(self.ts_end) < index + 1:
            self._ts_current += self.ts_end[index+1] - time.perf_counter()

    def _get_compare_val(self, instance: 'ChronoTimer'):
        val = ''
        if instance._dt_start:
            val += instance.get_str_dt_start

        if instance._dt_end:
            val += ' {}'.format(instance.get_str_dt_end)

        if instance.name:
            val += ' {}'.format(instance.name)

        return val